# Copyright (C) 1999, 2000 Jay Beale
# Copyright (C) 2001, 2002 Hewlett Packard
# Licensed under the GNU General Public License

package Bastille::DNS;
use lib "/usr/lib";

use Bastille::API;
@ENV="";
$ENV{PATH}="";
$ENV{CDPATH}=".";
$ENV{BASH_ENV}="";

######## TO DO: Convert the chown line to API functions, so we don't
#########       have to use GLOBAL_PREFIX

#######################################################################
##                             DNS/BIND/NAMED                        ##
#######################################################################

&chrootBIND;
&DeactivateNamed;


sub chrootBIND {
    my $todo_header=0;
    my $TODOText = "";
    if ( &getGlobalConfig("DNS","chrootbind") eq "Y" ) {
	&ActionLog("# sub chrootBIND\n");

	# Only chroot bind if we are running on a Red Hat / Mandrake / HP-UX 
        # system and we haven't already chroot'ed it...
	if($GLOBAL_DISTRO =~ "^HP-UX"){

          $what  = &getGlobal('BIN',"what");
          $named = &getGlobal('FILE',"named");
          $awk   = &getGlobal('BIN',"awk");
          $findversion="$what $named | $awk '\$1==\"named\" {print \$2}'";
          $bindversion=`$findversion`;
          chomp $bindversion;
          $bindversion =~ s/^(\d*)\..*$/$1/;

          if ($bindversion < 8) {
             $TODOText .= 
		 "\n---------------------------------\nChroot'ed Bind:\n" .
		 "---------------------------------\n".
		 "Bastille can only chroot Bind versions 8 or higher.\n" .
		 "Please install the latest version of Bind and rerun\n" .
		 "bastille -b.\n\n";
	     $todo_header=1;
          } 
	  elsif ($bindversion > 9) {
	      &ErrorLog("WARNING: Bastille has not been tested with Bind \n" .
			"versions later than 9. Attempting to continue anyway.");
          }
	  else {
	      
	      my $extrasteps=0;
	      
	      my $groups  = &getGlobal('BIN',"groups");
	      $nameduser = `$groups named 2>/dev/null`;
	      chomp($nameduser);
	      
	      my $jail = &getGlobal('BDIR',"jail");
	      
	      if ($nameduser !~ "named") {
		  
		  # create non-privileged user and group named
		  my $groupadd = 'PATH="/usr/bin"; ' . &getGlobal('BIN',"groupadd") . " named";
		  my $groupdel = 'PATH="/usr/bin"; ' . &getGlobal('BIN',"groupdel") . " named";

		  if (! &B_System($groupadd,$groupdel)) {
		      $extrasteps = 1;
		  }

		  my $useradd = 'PATH="/usr/bin"; ' . &getGlobal('BIN',"useradd") . 
		      " -g named -d $jail/bind/tmp -s /usr/bin/false named"; 
		  my $userdel = 'PATH="/usr/bin"; ' . &getGlobal('BIN',"userdel") . " named";
		  
		  if (! &B_System($useradd, $userdel)) {
		      $extrasteps = 1;
		  }
	      }

	      # install generic chroot tree
	      if(&B_install_jail("bind",&getGlobal('BFILE',"jail.generic.hpux"))) {
		  my $success=0;
		  # add dev/null and dev/log to chroot tree
		  &B_mknod("", "$jail/bind/dev/null","c 3 0x000002");
		  &B_chmod(0666,"$jail/bind/dev/null");
		  &B_System(&getGlobal('BIN','mkfifo') . " -p -m 0666 " . $jail . "/bind/dev/log",
			    &getGlobal('BIN','rm') . " -f " . $jail . "/bind/dev/log\n" . 
			    &getGlobal('BIN',"echo") . " \"Your DNS is not running outside of it's\\n" . 
			    "chroot jail.  You must manually migrate your named server\\n" .
			    "back to your Named server's default defined location(s).\\n".
			    "After you have completed this, feel free to remove the jail directories\\n" .
			    "from your machine.  Be sure to reset NAMED_ARGS to equal \\\"-u daemon\\\"\\n" . 
			    "in /etc/rc.config.d/namesvrs and then restart the daemon using the init\\n" . 
			    "scripts.  Your bind jail directory is located in\\n" .
			    &getGlobal('BDIR',"jail") . "\\n\" >> " . &getGlobal('BFILE',"TOREVERT"));
		  
		  if(-e &getGlobal('BIN',"named-xfer")){
		      # install bind 8 specific files and directories.
		      &ActionLog("Bind 8 chroot directory setup\n");
		      if (&B_install_jail("bind", &getGlobal('BFILE',"jail.bind.hpux"))){
			  $success=1;
		      }
		  } 
		  elsif(-e &getGlobal('FILE',"named")){
		      # install bind 9 specific files and directories.
		      &ActionLog("Bind 9 chroot directory setup\n");
		      if(&B_install_jail("bind",&getGlobal('BFILE',"jail.bind9.hpux"))){
			  $success=1;
		      }
		  } 
		  else {
		      &ErrorLog("Bind does not appear to be on this system; no action was taken.\n");
		  }
		  
		  if ($success) {
		      &B_System(&getGlobal('BIN',"cat") . " " .&getGlobal('FILE',"group") . " | " .
				&getGlobal('BIN',"grep") . " named > $jail/bind/etc/group",&getGlobal('BIN',"rm")  . " $jail/bind/etc/group");
		      
		      &B_chmod(0444, "$jail/bind/etc/group");
		      &B_chown((getpwnam("bin"))[2], "$jail/bind/etc/group");
		      &B_chgrp((getgrnam("bin"))[2], "$jail/bind/etc/group");
		      
		      # move the pidfile into the jail and get the permissions 
		      # and ownership right
		      my $pidfile = &getGlobal('FILE', "named.pid");
		      
		      if ( -e $pidfile ) {
			  if ( ! -l $pidfile ) {
			      &B_cp($pidfile, "$jail/bind$pidfile");
			  }
		      } 
		      else {
			  &B_create_file("$jail/bind$pidfile");
		      }
		      &B_chown((getpwnam("named"))[2], "$jail/bind$pidfile");
		      &B_chgrp((getgrnam("named"))[2], "$jail/bind$pidfile");
		      &B_chmod(0644, "$jail/bind$pidfile");
		      
		      my $ch_rc = &getGlobal('BIN',"ch_rc");
		      my $initscript = &getGlobal('FILE',"chkconfig_named");
		      my $syslogd = &getGlobal('BIN',"syslogd");
		      #--80 chars wide----------------------------------------------------------------
		      if ($todo_header!=1){
			  $TODOText .= 
			      "\n---------------------------------\nChroot'ed Bind:\n" . 
			      "---------------------------------\n";
		      }
		      $TODOText .= 
			  "Bastille has created a simple \"named\" directory structure in\n" .
			  "$jail/bind/\n\n" .
			  "You need to take the following steps to configure your name server\n" .
			  "to run without root privileges in a chroot jail.  If you were not\n" .
			  "running named before and do not plan to in the future, then you can\n" .
			  "ignore these steps.\n\n";
		      
		      if ($extrasteps==1) {
			  $TODOText .= 
			      "First, there were errors recorded in the error log which you will \n" .
			      "need to resolve.  Depending on the severity of the errors, you may \n" .
			      "or may not be able to continue with the following steps.  The most\n" .
			      "likely error is that a \"named\" user or group already exists, in which\n" .
			      "case the following steps should work just fine and \"named\" will run as\n" .
			      "that user.\n\n";
		      }
		      
		      # create the finish-up script
		      my $sh = &getGlobal('BIN',"sh");
		      my $finishscript = &getGlobal('BFILE', "finish-named-chroot.sh");
		      &B_create_file("$finishscript");
		      &B_blank_file("$finishscript",'a$b This pattern does not match anything');
		      
		      &B_append_line("$finishscript","", "#!$sh\n");
		      &B_append_line("$finishscript","", "$initscript stop\n");
		      &B_append_line("$finishscript","",
				     "# Link the process id file into the jail and instruct the \"named\" init \n" .
				     "# script to chroot into the new environment and run as user \"named\".\n");
		      &B_append_line("$finishscript","",
				     "$ch_rc" . " -a -p NAMED_ARGS=" .
				     "\'\"\`(rm $pidfile; ln -s $jail/bind$pidfile $pidfile) 2>/dev/null\` " .
				     "-u named -t $jail/bind\"\'\n");
		      &B_append_line("$finishscript","",
				     "$ch_rc -a -p NAMED=1 # to enable named, if not already\n");
		      &B_append_line("$finishscript","", "$initscript start # start named using its init script\n\n");
		      &B_append_line("$finishscript","", &getGlobal('BIN',"sleep") . " 5 # sleep for 5 seconds\n\n");
		      &B_append_line("$finishscript","",
				     "if [ \"\$(ps -ef | grep named | grep jail)\" = \"\" ]; then\n" .
				     "echo named chroot was UNSUCCESSFUL.\n" .
				     "echo \"Check your syslog for clues.  (usually /var/adm/syslog/syslog.log)\"\n" .
				     "echo Also, you might check the permissions of the directories/files in\n" .
				     "echo $jail/bind\n" .
				     "else\n" .
				     "echo named appears to be running inside of the chroot jail\n" .
				     "echo \n" .
				     "fi\n");
		      
		      $TODOText .= 
			  "1: Stop the currently running \"named\" process using the system\'s\n" .
			  "   init script for \"named\" \n\n" .
			  
			  "   type_this> $initscript stop\n\n" .
			  
			  "   and remove all \"/var/run/named.*\" files if they exist.\n\n" .
			  
			  "   type_this> rm -r /var/run/named.*\n\n" .
			  
			  "2: Next, move all of the configuration files that the current \n" .
			  "   \"named\" implementation requires from their current locations into\n" .
			  "   the equivalent place in the jail structure.  This is at least a\n" .
			  "   \"named.conf\" file.  If there is a directory path specified in the\n" .
			  "   options section of the \"named.conf,\" that path must be made as well\n" .
			  "   as all of the files that are referenced from that directory.  Hint: \n" .
			  "   If you create links like: \"ln -s <in_jail_location> <out_of_jail_location>,\"\n" .
			  "   then you can unchroot named without breaking anything.  Remember, however,\n" .
			  "   that any files owned by the named user can be modified if an attacker\n" .
			  "   breaks into bind as the \"named\" user.  By creating links, an attacker\n" .
			  "   could make changes which affect files outside of the jail too.\n\n" . 
			  
			  "   NOTE: all bind database files need to be readable by the user \"named.\"\n\n" .
			  
			  "3: Run the finish-named-chroot.sh script that Bastille has created for you.\n\n" .
			  
			  "   type_this> $finishscript\n\n";
		      
		  }
		  else {
		      &ErrorLog("Error trying to install the chroot bind jail.\n");
		  }
	      }
	  }
	  
	  &B_TODO($TODOText);
      }
	else {
	    
	    if ( ($GLOBAL_DISTRO =~ /^RH/) or ($GLOBAL_DISTRO =~ /^MN/) ) {
		
		#
		# First figure out where the chroot prison would be setup, so that we can
		# check for its existence.  If it already exists, we won't recreate it.
		#

		my $user ="";
		my $group = "";
		my $gid = "";

		# Look for a DNS user already on the operating system
		foreach $name ( 'dns','named' ) {
		    if (getpwnam($name)) {
			$user = $name;
		    }
		}
		if ($user) {
		    $home_dns = (getpwnam($user))[7];
		}
		else {
		    $home_dns = '/home/dns';	 
		}

		# Continue only if we haven't already performed this chroot.

		unless ( -e "${GLOBAL+PREFIX}${home_dns}/lib/libc.so.6" ) {

		    # Make the dns chroot directory
		    &B_create_dir ( $home_dns );  		
		    
		    # Look for a DNS group already on the operating system
		    foreach $name ( 'dns','named' ) {
			if (getgrnam($name)) {
			    $group = $name;
			    $gid = getgrnam($name);
			}
		    }
		    
		    
		    #
		    # Create the dns user/group, if one doesn't yet exist
		    #
		    
		    unless ($group) {
			$gid = '53';
			&B_append_line ("/etc/group",":53:","dns:x:53:\n");
			$group = 'dns';
		    }
		    unless ($user) {		    
			&B_append_line ("/etc/passwd",":$gid:","dns:x:53:$gid::$home_dns:/bin/false\n");
			&B_append_line("/etc/shadow","^dns:","dns:*:11089:0:99999:7:::\n");
			$user = 'dns';
		    }
		    
		    # Populate the dns chroot directory
		    &B_create_dir ("$home_dns");
		    &B_create_dir ("$home_dns/etc");
		    &B_chmod (0755,"$home_dns/etc");
		    &B_create_dir ("$home_dns/lib");
		    &B_chmod (0755,"$home_dns/lib");
		    &B_create_dir ("$home_dns/dev");
		    &B_chmod (0755,"$home_dns/dev");
		    &B_create_dir ("$home_dns/usr");
		    &B_chmod (0755,"$home_dns/usr");
		    &B_create_dir ("$home_dns/usr/sbin");
		    &B_chmod (0755,"$home_dns/usr/sbin");
		    &B_create_dir ("$home_dns/var");
		    &B_chmod (0755,"$home_dns/var");
		    &B_create_dir ("$home_dns/var/named");
		    &B_chmod (0755,"$home_dns/var/named");
		    &B_create_dir ("$home_dns/var/run");
		    &B_chmod (0755,"$home_dns/var/run");
		    &B_create_dir ("$home_dns/var/run/named");
		    &B_chmod (0755,"$home_dns/var/run/named");
		    
		    
		    unless ($GLOBAL_LOGONLY) {
			&B_mknod(" -m 666 ","$home_dns/dev/null"," c 1 3");
			
			&B_cp("/etc/named.conf","$home_dns/etc/named.conf");
			
			# Should we tell them to move their name logs over?
			
			&B_cp("/var/named/named.ca","$home_dns/var/named/named.ca");
			&B_cp("/var/named/named.local","$home_dns/var/named/named.local");
			
			my $command=&getGlobal('BIN',"chown");
			`$command -R $user.$user ${GLOBAL_PREFIX}$home_dns/var/named $home_dns/var/run $home_dns/var/run/named`;
			&B_cp("/usr/sbin/named","$home_dns/usr/sbin/named");
			&B_chmod(0755,"$home_dns/usr/sbin/named");
			# Don't delete named -- let it get loaded from /usr/sbin, but
			# find data files in the chroot directory.  We'll still put 
			# a copy of the binary in the chroot directory so the named
			# binary can be reloaded from there if necessary.
			#
			#&B_delete_file("/usr/sbin/named");
			my $namedxfer = &getGlobal('BIN',"named-xfer");
			if (-e $namedxfer ) {
			    &B_cp($namedxfer,"$home_dns" . $namedxfer);
			    &B_chmod(0755,"$home_dns" . $namedxfer);
			}
			&B_cp("/lib/libc.so.6","$home_dns/lib/libc.so.6");
			&B_cp("/lib/ld-linux.so.2","$home_dns/lib/ld-linux.so.2");
		    }
		    
		    # named communicates normally with syslog via the device /dev/log, 
		    # which isn't accessible from the chroot'd environment.  We set 
		    # syslog to create and listen to a specific device just for bind.
		    
		    &B_replace_line("/etc/rc.d/init.d/syslog",'daemon syslogd -m 0\s*\$',"daemon syslogd -m 0 -a $home_dns/dev/log\n");
		    &B_replace_line("/etc/rc.d/init.d/syslog",'daemon syslogd $SYSLOGD_OPTIONS\s*\$',"daemon syslogd \$SYSLOGD_OPTIONS -a $home_dns/dev/log\n");
		    
		    
		    # Modify named's init script to use the chroot environment
		    
		    my $bind8=1;	
		    if ($GLOBAL_DISTRO =~ /MN(\d+\.\d+)/) {
			if ($1 >= 8.0) {
			    $bind8=0;
			}
		    }
		    elsif ($GLOBAL_DISTRO =~ /RH(\d+\.\d+)/) {
                        if ($1 >= 7.1) {
			    $bind8=0;
			}	 
		    }
	
		    if ( $bind8 ) {
			&B_replace_line("/etc/rc.d/init.d/named",'^\s*daemon\s+named',"daemon named -u $user -g $group -t $home_dns\n");
		    }
		    else {
			# Mandrake 8.0 and later, along with Red Hat 7.1 and later, use BIND 9, which deprecated the -g option.
			# Nicely, all of these run as an alternate user already.
			
			if ($GLOBAL_DISTRO =~ /MN/) {
			    &B_replace_line("/etc/rc.d/init.d/named",'^\s*daemon\s+named',"daemon named -u $user -t $home_dns\n");  
			}
			elsif ($GLOBAL_DISTRO =~ /RH/) {
			    &B_append_line('/etc/sysconfig/named','^\s*ROOTDIR\s*\=\s*',"ROOTDIR=$home_dns\n");
			}
		    }
		    
		}
		else {
		    # We only do this to Red Hat and Mandrake boxes for now, because
		    # we'd have to find all the libraries and modify startup files
		    # and this can just be totally difficult to automate when you
		    # don't know the target...
		    
		    &ErrorLog("Couldn't chroot named! Not on a Red Hat / Mandrake box or $home_dns already exists.\n");
		}
		
	    } 
	}
    }
}

sub DeactivateNamed { 


    if (&getGlobalConfig("DNS","namedoff") eq "Y") {

	&ActionLog("# sub DeactivateNamed\n");

	# Deactivate BIND unless it is being used.

	&B_chkconfig_off ("named");
    }

} 


1;





